#!/usr/bin/python
##################################################################
##                                                              ##
## Author:     Louis-Philippe Lemieux Perreault                 ##
## Date:       28-10-2008                                       ##
## Source:     cn_genotype2pedfile.py                           ##
##                                                              ##
## Function: Convert CNGen results file to a pre-linkage        ##
##           format for a pedcheck run.                         ##
##                                                              ##
##################################################################

import os, sys

from os import path
from optparse import OptionParser


# The 'OptionParser' object and usage
parser = OptionParser()
parser.set_usage("Usage: python " + sys.argv[0] + " -i input_file -p pedigree_file -o output_file")

__encoding = {}
__all_family_structure = {}
__all_individual_ID = []
__nb_allele = -9999

def main():

	inputFileName, pedFileName, outputFileName = parseOptions()

	read_pedFile(open(pedFileName, 'rb'))

	inFile = open(inputFileName, 'rb')

	# Reads the first line of the input file in order to get the sample IDs
	line = inFile.readline()
	sampleID_list = line.strip().split('\t')[1:]
	
	sampleID_set = set()
	for sampleID in sampleID_list:
		sampleID_set.add(sampleID)
	
	# Creates the temporary files and find the most commun CN genotypes (for encoding)
	create_temp_files(inFile, sampleID_list)
	
	# Closes the input file
	inFile.close()

	create_output_file(open(outputFileName, 'wb'), sampleID_list, sampleID_set)
	


def create_output_file(outputFile, sampleID_list, sampleID_set):

	global __encoding
	global __all_family_structure
	global __all_individual_ID
	global __nb_allele
	
	not_genotype_individual_line_to_write = ['0' for i in range(__nb_allele)]

	files = []
	for famID, sampleID in __all_individual_ID:
		# For all sampleIDs (genotyped or not genotyped)
		line_to_write = []
		line_to_write.extend(__all_family_structure[(famID, sampleID)].get_pedcheck_info())
		
		if sampleID in sampleID_set:
			# If the sampleID has been genotyped
			# Opens the temp file
			temp_file = open(path.join('tmp', sampleID), 'rb')
			
			# We read the information from disk
			data = temp_file.readline().strip().split(' ')[1:]
			
			for cn_genotype in data:
				allele = cn_genotype
				line_to_write.append(__encoding[allele])

			# We close the temp file
			temp_file.close()
			
		else:
			# The sampleID has not been genotyped
			line_to_write.extend(not_genotype_individual_line_to_write)
			
		outputFile.write(' '.join(line_to_write))
		outputFile.write('\n')

	outputFile.close()


def create_temp_files(inputFile, sampleID_list):

	global __encoding
	global __nb_allele
	
	allele_counting = {}
	
	nb_marker = 0

	# Lets create a temporary directory
	os.mkdir('tmp')
	
	# Lets create some temporary files
	files = []
	for sampleID in sampleID_list:
		files.append(open(path.join('tmp', sampleID), 'wb'))
	
	for i, sampleID in enumerate(sampleID_list):
		files[i].write(sampleID)
		
	line = inputFile.readline()
	while line != '':
		nb_marker += 1
		data = line.strip().split('\t')[1:]
		
		for i in range(len(sampleID_list)):
			allele1, allele2 = data[i].split(',')
			
			if allele1 not in allele_counting:
				allele_counting[allele1] = 1
			else:
				allele_counting[allele1] = allele_counting[allele1] + 1
			
			if allele2 not in allele_counting:
				allele_counting[allele2] = 1
			else:
				allele_counting[allele2] = allele_counting[allele2] + 1
				
			files[i].write(' ' + ' '.join([allele1, allele2]))
		
		line = inputFile.readline()
	
	# Lets close all the temporary files
	for f in files:
		f.close()
	
	# Sets the encoding of the 2 undefined cn genotypes (-1 and-2)
	__encoding['-1'] = '0'
	__encoding['-2'] = '0'
	
	cn_genotypes = []
	for key, value in allele_counting.iteritems():
		if key != '-1' and key != '-2':
			cn_genotypes.append((value, key))
	cn_genotypes.sort(reverse = True)
	
	print "\nDone reading", nb_marker, "markers (" + str(nb_marker * 2), "alleles) for", str(len(sampleID_list)), "genotyped individuals"
	
	print "\nThis is the encoding for the output file: (" + str(len(cn_genotypes)), "different cn alleles)"
	for index, value in enumerate(cn_genotypes):
		key = value[1]
		rank = index + 1
		__encoding[key] = str(rank)
		print "\t" + key + ':\t', rank
	
	__nb_allele = nb_marker * 2
			


##
## Prints an ERROR message and exit with code 1
##		
def usage(message):

	print >>sys.stderr, "\nERROR: " + message + '\n'
	parser.print_help()
	sys.exit(1)



##
## Parse the options and arguments
##
def parseOptions():

	# Required arguments
	parser.add_option("-i", "--input", dest = "inputFileName",
						help = "The input FILE (cn genotypes format)",
						metavar = "FILE")
	
	parser.add_option("-p", "--ped", dest = "pedFileName",
						help = "The pedfile FILE name",
						metavar = "FILE")
	
	parser.add_option("-o", "--output", dest = "outputFileName", default = 'pedcheck_ready',
						help = "The output FILE",
						metavar = "FILE")
	
	# Parse the option
	options, args = parser.parse_args()
	
	# Gets the options
	inputFileName = options.inputFileName
	pedFileName = options.pedFileName
	outputFileName = options.outputFileName
	
	# Check if the user specified a input file
	if inputFileName == None:
		usage("You must specify the input file ('larry_bird_result FILE')")
	if pedFileName == None:
		usage("You must specify the pedfile name")
	
	return inputFileName, pedFileName, outputFileName



def read_pedFile(pedFile):

	global __all_family_structure
	global __all_individual_ID

	for line in pedFile:
		data = line.strip().split(' ')
		
		currentID = data[1]
		famID = data[0]
		
		temp_Individual = Individual(currentID)
		
		temp_Individual.set_family_ID(famID)	# The family ID
		temp_Individual.set_father_ID(data[2])	# The father ID
		temp_Individual.set_mother_ID(data[3])	# The mother ID
		temp_Individual.set_sex(data[4])		# The sex
		temp_Individual.set_affection(data[5])	# The affection status
		
		__all_family_structure[(famID, currentID)] = temp_Individual
		__all_individual_ID.append((famID, currentID))
	
	pedFile.close
	
	print "\nDone reading pedfile (" + str(len(__all_individual_ID)), "individuals)"



##################################################################################################
##
## The Individual class
##
## The Individual class contains all the required informations about
## an individual. It contains:
##    - family ID
##    - current ID
##    - father ID
##    - mother ID
##    - sex
##    - affection status
##
class Individual:

	__family_ID = None		# The family ID
	__current_ID = None		# The ID of the individual
	__father_ID = None		# The ID of the father
	__mother_ID = None		# The ID of the mother
	__sex = None			# The sex of the individual
	__affection = None		# The affection status of the individual

	##
	## The '__init__' method of the Individual class
	##
	## This is the structure of an individual.
	##
	def __init__(self, subjectID):
		
		# Initialize the ID of the individual
		self.__current_ID = subjectID
	
	
	
	##
	## The '__str__' method of the Individual class
	## 
	## Return the ID of the individual
	##
	def __str__(self):
	
		return self.__current_ID
	
	
	
	##
	## The 'get_ID' method of the Individual class
	##
	## Return the ID of the individual
	##
	def get_ID(self):
		
		return self.__current_ID
	
	
	
	def set_family_ID(self, famID):
		
		self.__family_ID = famID
		
		
	
	def get_family_ID(self):
		
		return self.__family_ID
		
		
		
	def set_father_ID(self, father):
		
		self.__father_ID = father
	
	
	
	def get_father_ID(self):
		
		return self.__father_ID
	
	
	
	def set_mother_ID(self, mother):
	
		self.__mother_ID = mother
	
	
	
	def get_mother_ID(self):
	
		return self.__mother_ID
	
	
	
	def set_sex(self, sex):
	
		self.__sex = sex
	
	
	
	def get_sex(self):
		
		return self.__sex
	
	
	
	def set_affection(self, affection_status):
	
		self.__affection = affection_status
	
	
	
	def get_affection(self):
		
		return self.__affection
	
	
	
	def get_pedcheck_info(self):
	
		return [self.__family_ID, self.__current_ID, self.__father_ID, self.__mother_ID, self.__sex]#, self.__affection]
	
##################################################################################################
		
	

if __name__ == '__main__':
	try:
		main()
	finally:
		# Lets remove temp files and tmp directory
		for root, dirs, files in os.walk('tmp', topdown=False):
			for name in files:
				os.remove(path.join(root, name))
			for name in dirs:
				os.rmdir(path.join(root, name))
		if os.path.exists('tmp'):
			os.rmdir('tmp')

